import gc
from collections import defaultdict
from io import BytesIO
from typing import Optional

import keras_ocr
import pytesseract
import tensorflow as tf
from jina import DocumentArray, Executor, requests
from PIL import Image

from now.constants import TAG_OCR_DETECTOR_TEXT_IN_DOC

keras_ocr.config.configure()


class NOWOCRDetector9(Executor):
    """Uses keras-ocr to detect text in images"""

    def __init__(self, traversal_paths: Optional[str] = '@r', *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.traversal_paths = traversal_paths
        self.keras_ocr_pipeline = keras_ocr.pipeline.Pipeline()
        # we need this to occasionally reset the keras-ocr pipeline to avoid memory leaks due to tensorflow
        self.cnt_images_inference = 0

    @requests(on=['/detect_text', '/index'])
    def detect_text(self, docs: DocumentArray, parameters: dict = {}, **kwargs):
        traversal_paths = parameters.get('traversal_paths', self.traversal_paths)
        flat_docs = docs[traversal_paths]

        if len(flat_docs) == 0:
            return

        id_to_text = defaultdict(str)
        # select documents whose mime_type starts with 'image'
        flat_docs = [doc for doc in flat_docs if doc.mime_type.startswith('image')]
        # use pytesseract to detect text in images
        for doc in flat_docs:
            text_in_doc = (
                pytesseract.image_to_string(Image.open(BytesIO(doc.blob)))
                .lower()
                .strip()
            )
            if traversal_paths == '@c':
                id_to_text[doc.parent_id] += text_in_doc + ' '
            else:
                id_to_text[doc.id] += text_in_doc + ' '
        # use keras-ocr to detect text in images
        images = [keras_ocr.tools.read(BytesIO(doc.blob)) for doc in flat_docs]
        self.cnt_images_inference += len(images)
        if images:
            predictions_keras = self.keras_ocr_pipeline.recognize(images)
            for doc, prediction in zip(flat_docs, predictions_keras):
                text_in_doc = ' '.join([word[0] for word in prediction])
                if traversal_paths == '@c':
                    id_to_text[doc.parent_id] += text_in_doc + ' '
                else:
                    id_to_text[doc.id] += text_in_doc + ' '
        # update the tags of the documents to include the detected text
        for doc in flat_docs:
            text_in_doc = id_to_text[
                doc.parent_id if traversal_paths == '@c' else doc.id
            ]
            doc.tags[TAG_OCR_DETECTOR_TEXT_IN_DOC] = text_in_doc.strip()
        # reset the keras-ocr pipeline every 100 images to avoid memory leaks
        if self.cnt_images_inference > 100:
            self.keras_ocr_pipeline = None
            # call garbage collector to free memory
            gc.collect()
            # if using gpu, call tensorflow to free memory
            tf.keras.backend.clear_session()
            self.keras_ocr_pipeline = keras_ocr.pipeline.Pipeline()
            self.cnt_images_inference = 0



import base64
import io
import os
from typing import Dict, List

import numpy as np
import PIL
from docarray import Document, DocumentArray
from jina.serve.runtimes.gateway.http.models import JinaRequestModel, JinaResponseModel
from pydantic import BaseModel

from deployment.bff.app.v1.models.text import NowTextSearchRequestModel
from deployment.bff.app.v1.models.video import (
    NowVideoIndexRequestModel,
    NowVideoListResponseModel,
    NowVideoResponseModel,
)
from now.app.base.app import JinaNOWApp
from now.common.utils import (
    _get_clip_apps_with_dict,
    common_setup,
    get_email,
    get_indexer_config,
)
from now.constants import CLIP_USES, Apps, Modalities
from now.now_dataclasses import UserInput

NUM_FRAMES_SAMPLED = 3


class TextToVideo(JinaNOWApp):
    def __init__(self):
        super().__init__()

    @property
    def app_name(self) -> str:
        return Apps.TEXT_TO_VIDEO

    @property
    def is_enabled(self) -> bool:
        return True

    @property
    def description(self) -> str:
        return 'Text to video search app'

    @property
    def input_modality(self) -> Modalities:
        return Modalities.TEXT

    @property
    def output_modality(self) -> Modalities:
        return Modalities.VIDEO

    @property
    def required_docker_memory_in_gb(self) -> int:
        return 12

    def set_flow_yaml(self, **kwargs):
        finetuning = kwargs.get('finetuning', False)
        dataset_len = kwargs.get('dataset_len', 0) * NUM_FRAMES_SAMPLED
        is_jina_email = get_email().split('@')[-1] == 'jina.ai'

        flow_dir = os.path.abspath(os.path.join(__file__, '..'))

        if finetuning:
            if dataset_len > 200_000 and is_jina_email:
                print(f"🚀🚀🚀 You are using high performance flow")
                self.flow_yaml = os.path.join(
                    flow_dir, 'ft-flow-video-clip-high-performance.yml'
                )
            else:
                self.flow_yaml = os.path.join(flow_dir, 'ft-flow-video-clip.yml')
        else:
            if dataset_len > 200_000 and is_jina_email:
                print(f"🚀🚀🚀 You are using high performance flow")
                self.flow_yaml = os.path.join(
                    flow_dir, 'flow-video-clip-high-performance.yml'
                )
            else:
                self.flow_yaml = os.path.join(flow_dir, 'flow-video-clip.yml')

    @property
    def supported_file_types(self) -> List[str]:
        return ['gif', 'mp4', 'mov']

    @property
    def index_query_access_paths(self) -> str:
        return '@c'

    def setup(
        self, dataset: DocumentArray, user_input: UserInput, kubectl_path
    ) -> Dict:
        indexer_config = get_indexer_config(len(dataset) * NUM_FRAMES_SAMPLED)
        encoder_with, ocr_with = _get_clip_apps_with_dict(user_input)
        env_dict = common_setup(
            app_instance=self,
            user_input=user_input,
            dataset=dataset,
            encoder_uses=CLIP_USES[user_input.deployment_type][0],
            encoder_with=encoder_with,
            encoder_uses_with={
                'pretrained_model_name_or_path': CLIP_USES[user_input.deployment_type][
                    1
                ]
            },
            pre_trained_embedding_size=CLIP_USES[user_input.deployment_type][2],
            indexer_uses=indexer_config['indexer_uses'],
            kubectl_path=kubectl_path,
            indexer_resources=indexer_config['indexer_resources'],
        )
        env_dict.update(ocr_with)
        super().setup(dataset, user_input, kubectl_path)
        return env_dict

    def preprocess(
        self, da: DocumentArray, user_input: UserInput, is_indexing=False
    ) -> DocumentArray:
        if is_indexing:

            def convert_fn(d: Document):
                try:
                    if d.blob == b'':
                        if d.uri:
                            d.load_uri_to_blob(timeout=10)
                        elif d.tensor is not None:
                            d.convert_tensor_to_blob()
                    sample_video(d)
                except Exception as e:
                    print(f'Failed to process {d.id}, error: {e}')
                return d

            for d in da:
                convert_fn(d)

            return DocumentArray(d for d in da if d.blob != b'')
        else:

            def convert_fn(d: Document):
                d.chunks = DocumentArray(d for d in d.chunks if d.text)
                for chunk in d.chunks:
                    if chunk.text == 'loader':
                        chunk.text = 'loading'
                return d

            for d in da:
                convert_fn(d)

            return DocumentArray(d for d in da if d.chunks)

    @property
    def bff_mapping_fns(self):
        def search_text_to_video_request_mapping_fn(
            request: NowTextSearchRequestModel,
        ) -> JinaRequestModel:
            jina_request_model = JinaRequestModel()
            jina_request_model.data = [Document(chunks=[Document(text=request.text)])]
            jina_request_model.parameters = {
                'limit': request.limit,
                'api_key': request.api_key,
                'jwt': request.jwt,
            }
            return jina_request_model

        def search_video_response_mapping_fn(
            request: NowTextSearchRequestModel, response: JinaResponseModel
        ) -> List[NowVideoResponseModel]:
            docs = response.data
            limit = request.limit
            return docs[0].matches[:limit].to_dict()

        def index_text_to_video_request_mapping_fn(
            request: NowVideoIndexRequestModel,
        ) -> JinaRequestModel:
            index_docs = DocumentArray()
            for video, uri, tags in zip(request.videos, request.uris, request.tags):
                if bool(video) + bool(uri) != 1:
                    raise ValueError(
                        f'Can only set one value but have video={video}, uri={uri}'
                    )
                if video:
                    base64_bytes = video.encode('utf-8')
                    message = base64.decodebytes(base64_bytes)
                    index_docs.append(Document(blob=message, tags=tags))
                else:
                    index_docs.append(Document(uri=uri, tags=tags))
            return JinaRequestModel(data=index_docs)

        def no_response_mapping_fn(_: JinaResponseModel) -> BaseModel:
            return BaseModel()

        return {
            '/search': (
                NowTextSearchRequestModel,
                NowVideoListResponseModel,
                search_text_to_video_request_mapping_fn,
                search_video_response_mapping_fn,
            ),
            '/index': (
                NowVideoIndexRequestModel,
                BaseModel,
                index_text_to_video_request_mapping_fn,
                no_response_mapping_fn,
            ),
        }

    @property
    def max_request_size(self) -> int:
        """Max number of documents in one request"""
        return 2


def select_frames(num_selected_frames, num_total_frames):
    partition_size = num_total_frames / (num_selected_frames + 1)
    return [round(partition_size * (i + 1)) for i in range(num_selected_frames)]


def sample_video(d):
    video = d.blob
    video_io = io.BytesIO(video)
    gif = PIL.Image.open(video_io)
    frame_indices = select_frames(NUM_FRAMES_SAMPLED, gif.n_frames)
    frames = []
    for i in frame_indices:
        gif.seek(i)
        frame = np.array(gif.convert("RGB"))
        frame_pil = PIL.Image.fromarray(frame)
        frame_pil_resized = frame_pil.resize((224, 224))
        frames.append(frame_pil_resized)
        frame_bytes = io.BytesIO()
        frame_pil_resized.save(frame_bytes, format="JPEG", quality=70)
        d.chunks.append(Document(uri=d.uri, blob=frame_bytes.getvalue(), tags=d.tags))



from collections import defaultdict
from copy import deepcopy
from sys import maxsize
from typing import List, Optional

from docarray import Document, DocumentArray

from now.constants import TAG_INDEXER_DOC_HAS_TEXT, TAG_OCR_DETECTOR_TEXT_IN_DOC
from now.executor.abstract.auth import NOWAuthExecutor as Executor
from now.executor.abstract.auth import SecurityLevel, secure_request
from now.executor.abstract.base_indexer.ranking import merge_matches_sum

CLOUD_BUCKET_PREFIXES = ['s3://']


class NOWBaseIndexer(Executor):
    def __init__(
        self,
        dim: int,
        columns: Optional[List] = None,
        metric: str = 'cosine',
        limit: int = 10,
        traversal_paths: str = '@r',
        max_values_per_tag: int = 10,
        *args,
        **kwargs,
    ):
        """
        :param dim: Dimensionality of vectors to index
        :param columns: List of tuples of the form (column_name, str_type). Here str_type must be a string that can be
        parsed as a valid Python type.
        :param metric: Distance metric type. Can be 'euclidean', 'inner_product', or 'cosine'
        :param limit: Number of results to get for each query document in search
        :param traversal_paths: Default traversal paths on docs
        :param max_values_per_tag: Maximum number of values per tag
        (used for search), e.g. '@r', '@c', '@r,c'
        """

        super().__init__(*args, **kwargs)
        self.columns = self.parse_columns(columns)
        self.dim = dim
        self.metric = metric
        self.limit = limit
        self.traversal_paths = traversal_paths
        self.max_values_per_tag = max_values_per_tag
        self.construct()
        self.doc_id_tags = {}
        self.document_list = DocumentArray()
        self.load_document_list()

    @secure_request(on='/tags', level=SecurityLevel.USER)
    def get_tags_and_values(self, **kwargs):
        """Returns tags and their possible values

        for example if indexed docs are the following:
            docs = DocumentArray([
                Document(.., tags={'color':'red'}),
                Document(.., tags={'color':'blue'}),
                Document(.., tags={'greeting':'hello'}),
            ])

        the resulting response would be a document array with
        one document containing a dictionary in tags like the following:
        {'tags':{'color':['red', 'blue'], 'greeting':['hello']}}
        """

        count_dict = defaultdict(lambda: defaultdict(int))
        for tags in self.doc_id_tags.values():
            for key, value in tags.items():
                count_dict[key][value] += 1

        tag_to_values = dict()
        for key, value_to_count in count_dict.items():
            sorted_values = sorted(
                value_to_count.items(), reverse=True, key=lambda item: item[1]
            )
            tag_to_values[key] = [
                value
                for (value, _), _ in zip(sorted_values, range(self.max_values_per_tag))
            ]
        return DocumentArray([Document(text='tags', tags={'tags': tag_to_values})])

    @secure_request(on='/list', level=SecurityLevel.USER)
    def list(self, parameters: dict = {}, **kwargs):
        """List all indexed documents.
        :param parameters: dictionary with limit and offset
        - offset (int): number of documents to skip
        - limit (int): number of retrieved documents
        """
        limit = int(parameters.get('limit', maxsize))
        offset = int(parameters.get('offset', 0))
        # add removal of duplicates
        traversal_paths = parameters.get('traversal_paths', self.traversal_paths)
        if traversal_paths == '@c':
            docs = DocumentArray()
            chunks_size = int(parameters.get('chunks_size', 3))
            parent_ids = set()
            for d in self.document_list[offset * chunks_size :]:
                if len(parent_ids) == limit:
                    break
                if d.parent_id in parent_ids:
                    continue
                parent_ids.add(d.parent_id)
                docs.append(d)
        else:
            docs = self.document_list[offset : offset + limit]
        return docs

    @staticmethod
    def set_tags_if_text_in_doc_matching_title(docs: DocumentArray):
        # update the tags of the documents to include the detected text
        for doc in docs:
            text_in_doc = doc.tags.pop(TAG_OCR_DETECTOR_TEXT_IN_DOC, '')
            text_in_doc = text_in_doc.split(' ')
            text_in_doc = filter(lambda s: len(s) > 1 and s.isalnum(), text_in_doc)
            doc.tags[TAG_INDEXER_DOC_HAS_TEXT] = len(list(text_in_doc)) > 0

    @secure_request(on='/index', level=SecurityLevel.USER)
    def index_endpoint(
        self,
        docs: Optional[DocumentArray] = None,
        parameters: dict = {},
        docs_matrix: Optional[List[DocumentArray]] = None,
        **kwargs,
    ):
        """Base function for indexing documents. Handles the data management for the index and list endpoints.

        :param docs: the Documents to index
        :param parameters: dictionary with options for indexing
        """
        traversal_paths = parameters.get('traversal_paths', self.traversal_paths)
        # merge the docs_matrix into the flat_docs
        if (
            docs_matrix
            and len(docs_matrix) > 1
            and all(len(da) > 0 for da in docs_matrix)
        ):
            flat_docs = DocumentArray()
            for doc0, doc1 in zip(*[_da[traversal_paths] for _da in docs_matrix]):
                if doc0.embedding is not None:
                    doc_encoded = doc0
                    doc_text = doc1
                else:
                    doc_encoded = doc1
                    doc_text = doc0
                doc_encoded.tags[TAG_OCR_DETECTOR_TEXT_IN_DOC] = doc_text.pop(
                    TAG_OCR_DETECTOR_TEXT_IN_DOC, ''
                )
                flat_docs.append(doc_encoded)
        else:
            flat_docs = docs[traversal_paths]
        if len(flat_docs) == 0:
            return

        self.set_tags_if_text_in_doc_matching_title(flat_docs)

        flat_docs = self.maybe_drop_blob_tensor(flat_docs)
        self.index(flat_docs, parameters, **kwargs)
        self.extend_inmemory_docs_and_tags(flat_docs)
        return DocumentArray([])

    @secure_request(on='/delete', level=SecurityLevel.USER)
    def delete_endpoint(self, parameters: dict = {}, **kwargs):
        """
        Delete endpoint to delete document/documents from the index.
        Filter conditions can be passed to select documents for deletion.
        """
        filter = parameters.get("filter", {})
        if filter:
            filtered_docs = deepcopy(self.document_list.find(filter=filter))
            self.delete_inmemory_docs_and_tags(filtered_docs)
            self.delete(filtered_docs, parameters, **kwargs)

        return DocumentArray()

    @secure_request(on='/search', level=SecurityLevel.USER)
    def search_endpoint(
        self, docs: Optional[DocumentArray] = None, parameters: dict = {}, **kwargs
    ):
        """Perform a vector similarity search and retrieve Document matches"""
        limit = int(parameters.get('limit', self.limit))
        search_filter_raw = parameters.get('filter', {})
        search_filter_orig = deepcopy(search_filter_raw)
        traversal_paths = parameters.get('traversal_paths', self.traversal_paths)
        docs = docs[traversal_paths][:1]  # only search on the first document for now

        if traversal_paths == '@c':
            retrieval_limit = limit * 3
        else:
            retrieval_limit = limit

        # if OCR detector was used to check if documents contain text in indexed image modality adjust retrieval step
        if self.columns and TAG_INDEXER_DOC_HAS_TEXT in [
            col[0] for col in self.columns
        ]:
            # enhance filter for one word queries to include the title
            if (
                len(docs[0].text.split()) == 1
                and not search_filter_raw
                and 'title' in [col[0] for col in self.columns]
            ):
                search_filter_raw = {'title': {'$regex': docs[0].text.lower()}}
            # search for documents which don't contain text
            search_filter_raw[TAG_INDEXER_DOC_HAS_TEXT] = {'$eq': False}
            search_filter = self.convert_filter_syntax(search_filter_raw)
            docs_with_matches_no_text = self.create_matches(
                docs,
                parameters,
                traversal_paths,
                limit,
                retrieval_limit,
                search_filter,
            )
            # search for documents which contain text
            search_filter_raw[TAG_INDEXER_DOC_HAS_TEXT] = {'$eq': True}
            search_filter = self.convert_filter_syntax(search_filter_raw)
            docs_with_matches_filter_title = self.create_matches(
                docs,
                parameters,
                traversal_paths,
                limit,
                retrieval_limit,
                search_filter,
            )
            # merge the results such that documents with text are retrieved at a later stage
            self.merge_matches_by_score_after_half(
                docs_with_matches_no_text,
                docs_with_matches_filter_title,
                limit,
            )
            docs_with_matches = docs_with_matches_no_text
            # if we didn't retrieve enough results, try to fetch more
            if len(docs_with_matches) < limit:
                search_filter = self.convert_filter_syntax(search_filter_orig)
                docs_with_additonal_matches = self.create_matches(
                    docs,
                    parameters,
                    traversal_paths,
                    limit,
                    retrieval_limit,
                    search_filter,
                )
                self.append_matches_if_not_exists(
                    docs_with_matches, docs_with_additonal_matches, limit
                )
        else:
            search_filter = self.convert_filter_syntax(search_filter_orig)
            docs_with_matches = self.create_matches(
                docs, parameters, traversal_paths, limit, retrieval_limit, search_filter
            )

        self.clean_response(docs_with_matches)
        return docs_with_matches

    def create_matches(
        self, docs, parameters, traversal_paths, limit, retrieval_limit, search_filter
    ):
        docs_copy = deepcopy(docs)
        self.search(docs_copy, parameters, retrieval_limit, search_filter)
        if traversal_paths == '@c':
            merge_matches_sum(docs_copy, limit)
        return docs_copy

    def append_matches_if_not_exists(
        self, docs_with_matches, docs_with_matches_to_add, limit
    ):
        # get all parent_ids of the matches of the docs_with_matches
        parent_ids = set()
        for doc, doc_to_add in zip(docs_with_matches, docs_with_matches_to_add):
            for match in doc.matches:
                parent_ids.add(match.parent_id)

            # append matches to docs_with_matches if they are not already in the matches
            for match in doc_to_add.matches:
                if match.parent_id not in parent_ids:
                    if len(doc.matches) >= limit:
                        break
                    doc.matches.append(match)

    def merge_matches_by_score_after_half(
        self,
        docs_with_matches: DocumentArray,
        docs_with_matches_to_add: DocumentArray,
        limit: int,
    ):
        # get all parent_ids of the matches of the docs_with_matches
        parent_ids = set()
        for doc, doc_to_add in zip(docs_with_matches, docs_with_matches_to_add):
            score_name = (
                list(doc.matches[0].scores.keys())[0]
                if len(doc.matches) > 0
                else self.metric
            )
            for match in doc.matches[: limit // 2]:
                parent_ids.add(match.parent_id)

            possible_matches = deepcopy(doc.matches[limit // 2 :]) + doc_to_add.matches
            doc.matches = doc.matches[: limit // 2]
            ids_scores = [
                (match.id, match.scores[score_name].value) for match in possible_matches
            ]
            ids_scores.sort(key=lambda x: x[1], reverse='similarity' in score_name)
            for id, _ in ids_scores:
                _match = possible_matches[id]
                if _match.id not in parent_ids:
                    if len(doc.matches) >= limit:
                        break
                    doc.matches.append(_match)

    @staticmethod
    def maybe_drop_blob_tensor(docs: DocumentArray):
        """Drops `blob` or `tensor` from documents which have `uri` attribute set and
        whose 'uri' is a data-uri or is either in the cloud(S3) or can be loaded."""
        for doc in docs:
            if doc.uri:
                if doc.text:
                    continue
                else:
                    try:
                        if not doc.uri.startswith(f'data:{doc.mime_type}') and not any(
                            [
                                doc.uri.startswith(cloud_bucket_prefix)
                                for cloud_bucket_prefix in CLOUD_BUCKET_PREFIXES
                            ]
                        ):
                            doc.load_uri_to_blob(timeout=10)
                        doc.blob = None
                        doc.tensor = None
                        doc.mime_type = None
                    except Exception as e:  # noqa E722
                        continue
        return docs

    @staticmethod
    def clean_response(docs):
        """Removes embedding & OCR tags from the root level and also from the matches."""

        def _clean_response(doc: Document):
            """Inplace removes embedding & tags associated with OCR detection."""
            doc.embedding = None
            for _tag in [TAG_INDEXER_DOC_HAS_TEXT, TAG_OCR_DETECTOR_TEXT_IN_DOC]:
                doc.tags.pop(_tag, None)

        for doc in docs:
            _clean_response(doc)
            for match in doc.matches:
                _clean_response(match)

    @staticmethod
    def parse_columns(columns):
        """Parse the columns to index"""
        valid_input_columns = ['str', 'float', 'int', 'bool']
        if columns:
            corrected_list = []
            for i in range(0, len(columns), 2):
                corrected_list.append((columns[i], columns[i + 1]))
            columns = corrected_list
            for n, t in columns:
                assert (
                    t in valid_input_columns
                ), f'column of type={t} is not supported. Supported types are {valid_input_columns}'
        return columns

    def load_document_list(self):
        """is needed for the list endpoint"""
        document_list = DocumentArray()
        for batch in self.batch_iterator():
            self.extend_inmemory_docs_and_tags(batch)
        self.document_list = DocumentArray(
            sorted([d for d in document_list], key=lambda x: x.id)
        )

    def extend_inmemory_docs_and_tags(self, batch):
        """Extend the in-memory DocumentArray with new documents"""
        for d in batch:
            tags = deepcopy(d.tags)
            for _tag in [TAG_INDEXER_DOC_HAS_TEXT, TAG_OCR_DETECTOR_TEXT_IN_DOC]:
                tags.pop(_tag, None)
            self.document_list.append(
                Document(id=d.id, uri=d.uri, tags=tags, parent_id=d.parent_id)
            )
            self.doc_id_tags[d.id] = tags

    def delete_inmemory_docs_and_tags(self, docs):
        """Delete documents from the in-memory DocumentArray"""
        for d in docs:
            del self.document_list[d.id]
            self.doc_id_tags.pop(d.id)

    def update_inmemory_docs_and_tags(self, docs):
        """Update documents in the in-memory DocumentArray"""
        self.delete_inmemory_docs_and_tags(docs)
        self.extend_inmemory_docs_and_tags(docs)

    def construct(self, **kwargs):
        """Calls the constructor of the specialized indexer"""
        raise NotImplementedError

    def batch_iterator(self):
        """Needs to be implemented in derived classes. Iterates over all documents in batches and yields them"""
        raise NotImplementedError

    def convert_filter_syntax(self, search_filter={}, search_filter_not={}):
        """Needs to be implemented in derived classes. Converts the filter syntax to the syntax of the specific indexer"""
        raise NotImplementedError

    def index(self, docs: DocumentArray, parameters: dict, **kwargs):
        """Needs to be implemented in derived classes. Indexes the documents"""
        raise NotImplementedError

    def delete(self, docs_to_delete, parameters: dict, **kwargs):
        """Needs to be implemented in derived classes. Deletes the documents"""
        raise NotImplementedError

    def search(
        self,
        docs: DocumentArray,
        parameters: dict,
        limit: int,
        search_filter: dict,
        **kwargs,
    ):
        """Needs to be implemented in derived classes. Searches the documents"""
        raise NotImplementedError



from __future__ import annotations, print_function, unicode_literals

from now.utils import BetterEnum

# TODO: Uncomment the DEMO_DATASET_DOCARRAY_VERSION when the DocArray datasets on GCloud has been changed
# from docarray import __version__ as docarray_version

DEMO_DATASET_DOCARRAY_VERSION = '0.13.17'
NOW_PREPROCESSOR_VERSION = '0.0.95-add-tags-3'
DOCKER_BFF_PLAYGROUND_TAG = '0.0.130-test-multimodal-20'
NOW_QDRANT_INDEXER_VERSION = '0.0.3-feat-matching-text-82'
NOW_ELASTIC_INDEXER_VERSION = '0.0.3-matching-text-82'
NOW_AUTOCOMPLETE_VERSION = '0.0.1-feat-matching-text-82'
NOW_OCR_DETECTOR_VERSION = '0.0.1-feat-matching-text-82'


class Modalities(BetterEnum):
    TEXT = 'text'
    IMAGE = 'image'
    MUSIC = 'music'
    VIDEO = 'video'
    TEXT_AND_IMAGE = 'text-and-image'


class Apps(BetterEnum):
    TEXT_TO_TEXT = 'text_to_text'
    TEXT_TO_IMAGE = 'text_to_image'
    IMAGE_TO_TEXT = 'image_to_text'
    IMAGE_TO_IMAGE = 'image_to_image'
    MUSIC_TO_MUSIC = 'music_to_music'
    TEXT_TO_VIDEO = 'text_to_video'
    TEXT_TO_TEXT_AND_IMAGE = 'text_to_text_and_image'


class DatasetTypes(BetterEnum):
    DEMO = 'demo'
    PATH = 'path'
    URL = 'url'
    DOCARRAY = 'docarray'
    S3_BUCKET = 's3_bucket'
    ELASTICSEARCH = 'elasticsearch'


class Qualities(BetterEnum):
    MEDIUM = 'medium'
    GOOD = 'good'
    EXCELLENT = 'excellent'


class ModelNames(BetterEnum):
    MLP = 'mlp'
    SBERT = 'sentence-transformers/msmarco-distilbert-base-v3'
    CLIP = 'openai/clip-vit-base-patch32'


class ModelDimensions(BetterEnum):
    SBERT = 768
    CLIP = 512


BASE_STORAGE_URL = (
    'https://storage.googleapis.com/jina-fashion-data/data/one-line/datasets'
)

CLIP_USES = {
    'local': ('CLIPOnnxEncoder/latest', 'ViT-B-32::openai', ModelDimensions.CLIP),
    'remote': ('CLIPOnnxEncoder/latest-gpu', 'ViT-B-32::openai', ModelDimensions.CLIP),
}

EXTERNAL_CLIP_HOST = 'encoderclip-bh-5f4efaff13.wolf.jina.ai'
EXTERNAL_OCR_HOST = 'ocr-fb-55679da030.wolf.jina.ai'

PREFETCH_NR = 10

SURVEY_LINK = 'https://10sw1tcpld4.typeform.com/to/VTAyYRpR?utm_source=cli'

TAG_OCR_DETECTOR_TEXT_IN_DOC = '_ocr_detector_text_in_doc'
TAG_INDEXER_DOC_HAS_TEXT = '_indexer_doc_has_text'
